package main

import (
	"bytes"
	"context"
	"net"
	"net/http"
	"time"

	"gitlab.mashroom.com/mashroom/backend/identity/docs"

	"gitlab.mashroom.com/mashroom/backend/common/api/health"
	"gitlab.mashroom.com/mashroom/backend/common/generic/addr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/grpc"
	logger "gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/metrics"
	"gitlab.mashroom.com/mashroom/backend/common/generic/tracer"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	identity "gitlab.mashroom.com/mashroom/backend/identity/api"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/model/mongo"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/rentprofile"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/service"
	"gitlab.mashroom.com/mashroom/backend/identity/version"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
)

var (
	envMongoURL      = env.ResolveString("MONGO_URL", "mongodb:27017")
	envMongoUsername = env.ResolveString("MONGO_USERNAME", "extras-data")
	envMongoPassword = env.ResolveString("MONGO_PASSWORD", "extras-data")
	envMongoAuthDB   = env.ResolveString("MONGO_AUTH", "admin")
	envMongoDB       = env.ResolveString("MONGO_DATABASE", "extras-data")
)

var (
	connectDocumentService = document.NewDocumentServiceClient(grpc.Dial(
		env.ResolveString("DOCUMENT_SERVICE_URL", "mashroom-document:10000"),
		"Document",
	))

	connectOfferService = offer.NewOfferServiceClient(grpc.Dial(
		env.ResolveString("OFFER_SERVICE_URL", "mashroom-offer:10000"),
		"Offer",
	))
)

var (
	envSanctionListURL = env.ResolveString(
		"SANCTION_LIST_URL",
		"https://ofsistorage.blob.core.windows.net/publishlive/ConList.csv",
	)
	envRentprofileAPIKey  = env.ResolveStringPanic("RENTPROFILE_API_KEY")
	envEmulateRentprofile = env.ResolveBool("EMULATE_RENTPROFILE", false)
	envPollingInterval    = env.ResolveDuration("POLLING_INTERVAL", time.Minute*30)
)

func uploadTestDataFile(ctx context.Context, name, mime string) string {
	bs, err := docs.FS.ReadFile(name)
	if err != nil {
		panic(err)
	}

	f, err := document.UploadFile(
		ctx,
		connectDocumentService,
		name,
		mime,
		&document.FileAttribute{
			Data: map[string]string{
				document.FileAttributeDocumentType.String(): document.TypeOther,
			},
		},
		bytes.NewReader(bs),
	)
	if err != nil {
		panic(err)
	}

	return f.GetId()
}

func main() {
	ctx := context.Background()

	defer tracer.Close()

	repository, err := mongo.NewRepository(ctx, &mongo.Config{
		URL:        envMongoURL,
		Username:   envMongoUsername,
		Password:   envMongoPassword,
		Database:   envMongoDB,
		AuthSource: envMongoAuthDB,
	})
	if err != nil {
		panic(err)
	}

	err = repository.Migrate(ctx)
	if err != nil {
		panic(err)
	}

	rentprofileClient, err := rentprofile.New(&rentprofile.Config{
		HTTPClient: &http.Client{},
		Key:        envRentprofileAPIKey,
		URL:        rentprofile.DefaultURL,
	})
	if err != nil {
		panic(err)
	}

	var emulated service.Emulated

	if envEmulateRentprofile {
		emulated.Enabled = true
		emulated.HighRiskFileID = uploadTestDataFile(ctx, "highrisk.png", "image/png")
		emulated.MediumRiskFileID = uploadTestDataFile(ctx, "medrisk.png", "image/png")
		emulated.LowRiskFileID = uploadTestDataFile(ctx, "lowrisk.png", "image/png")
		emulated.TestPdfFileID = uploadTestDataFile(ctx, "example-tenant-check-report.pdf", "application/pdf")
	}

	srv, err := service.NewService(ctx, &service.Config{
		Repository: repository,
		Connections: service.Connections{
			Document: connectDocumentService,
			Offer:    connectOfferService,
		},
		Emulated:   emulated,
		Client:     rentprofileClient,
		HTTPClient: &http.Client{},
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("identity")
	if err != nil {
		panic(err)
	}

	identity.RegisterIdentityServiceServer(grpcServer, srv)

	logger.Infoln("identity started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	go func() {
		_, err = serveBPMS(srv)
		if err != nil {
			panic(err)
		}
	}()

	go func() {
		_ = metrics.ConfigureListenAndServe(grpcServer)
	}()

	go func() {
		err = srv.UpdateSanctionList(ctx, envSanctionListURL)
		if err != nil {
			logger.WithError(err).Error("fetching sanction list")
		}
	}()

	go func() {
		for {
			err = srv.PollStatusChanges(ctx, time.Now().Add(envPollingInterval))
			if err != nil {
				logger.WithError(err).Error("polling for status changes")
			}

			time.Sleep(envPollingInterval)
		}
	}()

	err = grpcServer.Serve(listener)
	if err != nil {
		logger.Errorln("gRPC error", err)
	}
}
