// Package mongo provides mongo implementation for persistent repository
package mongo

import (
	"context"
	"fmt"
	"time"

	"go.mongodb.org/mongo-driver/bson"
	"go.mongodb.org/mongo-driver/bson/primitive"
	"go.mongodb.org/mongo-driver/mongo"
	"go.mongodb.org/mongo-driver/mongo/options"

	"gitlab.mashroom.com/mashroom/backend/common/generic/migrations"
	migrationsMongo "gitlab.mashroom.com/mashroom/backend/common/generic/migrations/mongo"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/model"
	v1 "gitlab.mashroom.com/mashroom/backend/identity/internal/model/mongo/migrations/v1"
)

const (
	collectionTenant   = "tenant"
	collectionLandlord = "landlord"
	collectionSanction = "sanction"
)

// Config mongo repository config
type Config struct {
	URL        string
	Username   string
	Password   string
	Database   string
	AuthSource string
}

type repositoryImpl struct {
	client             *mongo.Client
	database           *mongo.Database
	collectionTenant   *mongo.Collection
	collectionLandlord *mongo.Collection
	collectionSanction *mongo.Collection
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid repository config")
	}

	if config.URL == "" {
		return fmt.Errorf("empty url in repository config")
	}

	if config.Username == "" {
		return fmt.Errorf("empty username in repository config")
	}

	if config.Password == "" {
		return fmt.Errorf("empty password in repository config")
	}

	if config.Database == "" {
		return fmt.Errorf("empty database in repository config")
	}

	if config.AuthSource == "" {
		return fmt.Errorf("empty database in repository config")
	}

	return nil
}

// NewRepository returns new repository instance
func NewRepository(ctx context.Context, config *Config) (model.Repository, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	client, err := mongo.NewClient(
		options.Client().ApplyURI(fmt.Sprintf("mongodb://%s", config.URL)).SetAuth(options.Credential{
			AuthSource:  config.AuthSource,
			Username:    config.Username,
			Password:    config.Password,
			PasswordSet: true,
		}),
	)
	if err != nil {
		return nil, err
	}

	err = client.Connect(ctx)
	if err != nil {
		return nil, err
	}

	db := client.Database(config.Database)

	return &repositoryImpl{
		client:             client,
		database:           db,
		collectionTenant:   db.Collection(collectionTenant),
		collectionLandlord: db.Collection(collectionLandlord),
		collectionSanction: db.Collection(collectionSanction),
	}, nil
}

func (repo *repositoryImpl) Migrate(ctx context.Context) error {
	err := migrations.MigrateSchema(migrationsMongo.WithInstance(ctx, repo.database),
		&migrationsMongo.Migration{
			Version: "v1",
			Up:      v1.Up,
		},
	)
	if err != nil {
		return err
	}

	return nil
}

func (repo *repositoryImpl) SanctionSave(ctx context.Context, sanction *model.Sanction) error {
	_, err := repo.collectionSanction.ReplaceOne(
		ctx,
		bson.M{
			"key": sanction.Key,
		},
		sanction,
		options.Replace().SetUpsert(true),
	)

	return err
}

func (repo *repositoryImpl) SanctionRemoveByMinVersion(ctx context.Context, min time.Time) error {
	_, err := repo.collectionSanction.DeleteMany(ctx, bson.M{
		"$expr": bson.M{
			"$lt": bson.A{"$version", min},
		},
	})

	return err
}

func (repo *repositoryImpl) SanctionCountByName(ctx context.Context, names ...string) (int, error) {
	cur, err := repo.collectionSanction.Aggregate(ctx, bson.A{
		// filter by index first
		bson.M{
			"$match": bson.M{
				"search": bson.M{
					"$in": names,
				},
			},
		},
		// $expr would not use multikey index
		bson.M{
			"$match": bson.M{
				"$expr": bson.M{
					"$setIsSubset": bson.A{
						names,
						"$search",
					},
				},
			},
		},
		bson.M{
			"$count": "count",
		},
	})
	if err != nil {
		return 0, err
	}

	var count int

	if cur.Next(ctx) {
		err = cur.Decode(&count)
		if err != nil {
			return 0, err
		}

		_ = cur.Close(ctx)
	}

	return count, nil
}

func (repo *repositoryImpl) pollingCollection(typ model.PollingType) *mongo.Collection {
	switch typ {
	case model.PollingTypeTenant:
		return repo.collectionTenant
	case model.PollingTypeLandlord:
		return repo.collectionLandlord
	default:
		return nil
	}
}

func (repo *repositoryImpl) PollingInsert(
	ctx context.Context,
	typ model.PollingType,
	m *model.Polling,
) error {
	col := repo.pollingCollection(typ)
	if col == nil {
		return nil
	}

	_, err := col.InsertOne(ctx, m)

	m.Type = typ

	return err
}

func (repo *repositoryImpl) PollingGetByRequestID(
	ctx context.Context,
	requestID string,
) (*model.Polling, error) {
	m := &model.Polling{}

	err := repo.collectionTenant.FindOne(ctx, bson.M{"request_id": requestID}).Decode(m)
	m.Type = model.PollingTypeTenant

	if err == mongo.ErrNoDocuments {
		err = repo.collectionLandlord.FindOne(ctx, bson.M{"request_id": requestID}).Decode(m)
		m.Type = model.PollingTypeLandlord
	}

	if err != nil {
		return nil, err
	}

	return m, nil
}

func (repo *repositoryImpl) PollingGetByTypeRequestID(
	ctx context.Context,
	typ model.PollingType,
	requestID string,
) (*model.Polling, error) {
	col := repo.pollingCollection(typ)
	if col == nil {
		return nil, mongo.ErrNoDocuments
	}

	m := &model.Polling{}

	err := col.FindOne(ctx, bson.M{"request_id": requestID}).Decode(m)
	if err != nil {
		return nil, err
	}

	m.Type = typ

	return m, nil
}

func (repo *repositoryImpl) PollingListByTypeLastPolled(
	ctx context.Context,
	typ model.PollingType,
	max time.Time,
) (ms []*model.Polling, err error) {
	col := repo.pollingCollection(typ)
	if col == nil {
		return
	}

	cur, err := col.Find(ctx, bson.M{
		"last_polled_at": bson.M{"$lt": max},
		"emulated_data":  nil,
	})
	if err != nil {
		return
	}

	err = cur.All(ctx, &ms)
	if err != nil {
		return
	}

	for _, m := range ms {
		m.Type = typ
	}

	return
}

func (repo *repositoryImpl) PollingUpdate(
	ctx context.Context,
	polling *model.Polling,
) error {
	col := repo.pollingCollection(polling.Type)
	if col == nil {
		return nil
	}

	_, err := col.ReplaceOne(ctx, bson.M{"_id": polling.ID}, polling)

	return err
}

func (repo *repositoryImpl) PollingUpdateLastPolledByTypeID(
	ctx context.Context,
	typ model.PollingType,
	id primitive.ObjectID,
	t time.Time,
) error {
	col := repo.pollingCollection(typ)
	if col == nil {
		return nil
	}

	_, err := col.UpdateOne(ctx, bson.M{"_id": id}, bson.M{"$set": bson.M{"last_polled_at": t}})

	return err
}

func (repo *repositoryImpl) PollingRemoveByTypeID(
	ctx context.Context,
	typ model.PollingType,
	id primitive.ObjectID,
) error {
	col := repo.pollingCollection(typ)
	if col == nil {
		return nil
	}

	_, err := col.DeleteOne(ctx, bson.M{"_id": id})

	return err
}

func (repo *repositoryImpl) PollingRemoveByTypeOffer(
	ctx context.Context,
	typ model.PollingType,
	offer string,
) error {
	col := repo.pollingCollection(typ)
	if col == nil {
		return nil
	}

	_, err := col.DeleteOne(ctx, bson.M{"offer": offer})

	return err
}
