package rentprofile

import (
	"bytes"
	"context"
	"encoding/json"
	"errors"
	"fmt"
	"net/http"
	"strconv"
)

type clientImpl struct {
	Config
}

func (client *clientImpl) exchange(
	ctx context.Context,
	path string,
	body []byte,
	resp interface{},
) error {
	hreq, err := http.NewRequestWithContext(ctx, http.MethodPost, client.URL+path, bytes.NewReader(body))
	if err != nil {
		return err
	}

	hreq.Header.Set("content-type", "application/json")
	hreq.Header.Set("accept", "application/json")
	hreq.Header.Set("content-length", strconv.Itoa(len(body)))

	hresp, err := client.HTTPClient.Do(hreq)
	if err != nil {
		return err
	}

	defer func() {
		_ = hresp.Body.Close()
	}()

	if hresp.StatusCode/100 != 2 {
		return errors.New(hresp.Status)
	}

	err = json.NewDecoder(hresp.Body).Decode(resp)
	if err != nil {
		return err
	}

	return nil
}

func (client *clientImpl) Invite(ctx context.Context, req *InviteRequest) (*InviteResponse, error) {
	reqdata := map[string]string{
		"api_key":       client.Key,
		"address_line1": req.AddressLine1,
		"postcode":      req.Postcode,
		"total_rent":    fmt.Sprintf("£%d", req.TotalRent),
	}

	for i, a := range req.Applicants {
		var ik string

		if i > 0 {
			ik = "_" + strconv.Itoa(i+1)
		}

		reqdata["applicant_name"+ik] = a.Name
		reqdata["applicant_email"+ik] = a.Email
		reqdata["applicant_mobile"+ik] = a.Mobile

		if a.IsGuarantorFor != "" {
			reqdata["applicant_is_guarantor_for"+ik] = a.IsGuarantorFor
		}
	}

	body, err := json.Marshal(reqdata)
	if err != nil {
		return nil, err
	}

	resp := make(map[string]string)

	err = client.exchange(ctx, "invite/request", body, &resp)
	if err != nil {
		return nil, err
	}

	var apps []*InviteResponseApplicant

	for i := 0; i < len(req.Applicants); i++ {
		var ik string

		if i > 0 {
			ik = "_" + strconv.Itoa(i+1)
		}

		apps = append(apps, &InviteResponseApplicant{
			Name:      resp["applicant_name"+ik],
			RequestID: resp["request_id"+ik],
		})
	}

	return &InviteResponse{
		Applicants: apps,
	}, nil
}

func (client *clientImpl) Status(ctx context.Context, req *StatusRequest) (*StatusResponse, error) {
	req.APIKey = client.Key

	body, err := json.Marshal(req)
	if err != nil {
		return nil, err
	}

	resp := &StatusResponse{}

	err = client.exchange(ctx, "status/request", body, resp)
	if err != nil {
		return nil, err
	}

	return resp, nil
}

func (client *clientImpl) Result(ctx context.Context, req *ResultRequest) (*ResultResponse, error) {
	req.APIKey = client.Key

	body, err := json.Marshal(req)
	if err != nil {
		return nil, err
	}

	resp := &ResultResponse{}

	err = client.exchange(ctx, "result/request", body, resp)
	if err != nil {
		return nil, err
	}

	return resp, nil
}
