// Package service provides implementation for identity service
package service

import (
	"context"
	"fmt"
	"net/http"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	identity "gitlab.mashroom.com/mashroom/backend/identity/api"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/model"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/rentprofile"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
)

var errEmulateModeIsActive = grpcerr.New(
	grpcerr.InvalidState,
	"emulate mode is active, not emulated entries are prohibited",
)

// Connections remote service connections
type Connections struct {
	Offer    offer.OfferServiceClient
	Document document.DocumentServiceClient
}

// Emulated configuration and responses
type Emulated struct {
	HighRiskFileID   string
	MediumRiskFileID string
	LowRiskFileID    string
	TestPdfFileID    string
	Enabled          bool
}

// Config for service implementation
type Config struct {
	// HTTP client for downloading documents
	HTTPClient *http.Client
	// remote API client
	Client rentprofile.Client
	// Persistent repository
	Repository model.Repository
	// Connections services connections
	Connections Connections
	// Emulated config
	Emulated Emulated
}

// Service implementation
type Service interface {
	identity.IdentityServiceServer
	UpdateSanctionList(ctx context.Context, url string) error
	PollStatusChanges(ctx context.Context, max time.Time) error
	SkipChecks(ctx context.Context, typ model.PollingType, offerID string) error
}

func validateConfig(config *Config) error {
	if config == nil {
		return fmt.Errorf("invalid service config")
	}

	if config.Repository == nil {
		return fmt.Errorf("no repository in service config")
	}

	if config.Connections.Offer == nil {
		return fmt.Errorf("no connection for offer client in service config")
	}

	if config.Connections.Document == nil {
		return fmt.Errorf("no connection for document client in service config")
	}

	if config.Client == nil {
		return fmt.Errorf("no rentprofile clinet in service config")
	}

	if config.Emulated.Enabled {
		if config.Emulated.LowRiskFileID == "" {
			return fmt.Errorf("no low risk file id for emulated response in service config")
		}

		if config.Emulated.MediumRiskFileID == "" {
			return fmt.Errorf("no medium risk file id for emulated response in service config")
		}

		if config.Emulated.HighRiskFileID == "" {
			return fmt.Errorf("no high risk file id for emulated response in service config")
		}

		if config.Emulated.TestPdfFileID == "" {
			return fmt.Errorf("no test pdf file id for emulated response in service config")
		}
	}

	return nil
}

// NewService returns new service instance
func NewService(ctx context.Context, config *Config) (Service, error) {
	err := validateConfig(config)
	if err != nil {
		return nil, err
	}

	srv := &serviceImpl{
		Config: *config,
	}

	return srv, nil
}
