package service

import (
	"context"
	"fmt"
	"io"
	"mime"
	"net/http"
	"os"
	"path"
	"strings"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	document "gitlab.mashroom.com/mashroom/backend/document/api"
	identity "gitlab.mashroom.com/mashroom/backend/identity/api"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/model"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/rentprofile"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
)

func translateRiskScore(s string) identity.TenantRisk {
	switch {
	case strings.Contains(s, "low"):
		return identity.TenantRisk_LOW_RISK
	case strings.Contains(s, "medium") || strings.Contains(s, "moderate"):
		return identity.TenantRisk_MEDIUM_RISK
	case strings.Contains(s, "high"):
		return identity.TenantRisk_HIGH_RISK
	default:
		return identity.TenantRisk_PENDING_RISK
	}
}

// checkUpdate sends Rentprofile requests to obtain checking progress for specified request ID.
// You should verify returning success response, it may not contain PDF file ID by design.
func (srv *serviceImpl) checkUpdate(
	ctx context.Context,
	requestID string,
	typ model.PollingType,
	pol *model.Polling,
) (err error) {
	logger := log.WithFields(log.Fields{
		"action": "checkUpdate",
		"poll": func() string {
			if pol != nil {
				return pol.ID.Hex()
			}
			return "false"
		}(),
		"request_id":   requestID,
		"polling_type": typ,
	})

	if pol != nil {
		err = srv.Repository.PollingUpdateLastPolledByTypeID(ctx, typ, pol.ID, time.Now())
		if err != nil {
			return err
		}
	}

	status, err := srv.Client.Status(ctx, &rentprofile.StatusRequest{
		RequestID: requestID,
	})
	if err != nil {
		return err
	}

	changedstatus, err := srv.processStatus(ctx, pol, status)
	if err != nil {
		logger.WithError(err).Error()
	}

	if pol.Status != rentprofile.TenantStatusComplete {
		return nil
	}

	res, err := srv.Client.Result(ctx, &rentprofile.ResultRequest{
		RequestID:   requestID,
		PDFRequired: rentprofile.BooleanTrue,
	})
	if err != nil {
		return err
	}

	changedresult, err := srv.processResult(ctx, pol, res)
	if err != nil {
		logger.WithError(err).Error()
	}

	err = srv.processChange(ctx, pol, changedstatus, changedresult)
	if err != nil {
		return err
	}

	return nil
}

func (srv *serviceImpl) processChange(
	ctx context.Context,
	pol *model.Polling,
	changedstatus, changedresult bool,
) (err error) {
	if changedstatus || changedresult || pol.Status == rentprofile.TenantStatusComplete {
		_, err = srv.Connections.Offer.TenantCheckUpdate(ctx, &offer.TenantCheckUpdateRequest{
			Offer:       pol.Offer,
			Check:       pol.RequestID,
			PdfId:       pol.PDFID,
			RiskScore:   pol.RiskScore,
			Status:      pol.Status,
			ProfileLink: pol.ProfileLink,
		})
		if err != nil {
			return err
		}
	}

	if pol.Status == rentprofile.TenantStatusComplete && pol.PDFID != "" {
		err = srv.Repository.PollingRemoveByTypeID(ctx, pol.Type, pol.ID)
		if err != nil {
			return err
		}
	}

	return
}

func (srv *serviceImpl) processStatus(
	ctx context.Context,
	pol *model.Polling,
	status *rentprofile.StatusResponse,
) (changed bool, err error) {
	changed = pol.Status != status.Status || pol.ProfileLink != status.ProfileLink

	pol.Status = status.Status
	pol.ProfileLink = status.ProfileLink

	if status.InviteDate != "" {
		var invt time.Time

		invt, err = time.Parse("02/01/2006", status.InviteDate)
		if err != nil {
			err = fmt.Errorf("parsing date %s: %w", status.InviteDate, err)
		}

		if pol.InviteDate != invt {
			pol.InviteDate = invt
			changed = true
		}
	}

	if changed {
		upderr := srv.Repository.PollingUpdate(ctx, pol)
		if upderr != nil {
			err = upderr
		}
	}

	return
}

func (srv *serviceImpl) processResult(
	ctx context.Context,
	pol *model.Polling,
	result *rentprofile.ResultResponse,
) (changed bool, err error) {
	changed = pol.PDFURL != result.PDFURL || pol.RiskScore != result.RiskScore

	if result.PDFURL != "" && (pol.PDFURL != result.PDFURL || pol.PDFID == "") {
		var fileID string

		fileID, err = srv.downloadPDF(ctx, result.PDFURL)
		if err != nil {
			err = fmt.Errorf("downloading document %s: %w", result.PDFURL, err)
		}

		if fileID != "" {
			pol.PDFID = fileID
			changed = true
		}
	}

	pol.PDFURL = result.PDFURL
	pol.RiskScore = result.RiskScore

	if changed {
		upderr := srv.Repository.PollingUpdate(ctx, pol)
		if upderr != nil {
			err = upderr
		}
	}

	return
}

// downloadPDF tries to download rentprofile pdf report and then upload it into document service.
// Returns file ID on success and error otherwise.
func (srv *serviceImpl) downloadPDF(ctx context.Context, url string) (string, error) {
	if srv.Emulated.Enabled {
		return srv.Emulated.TestPdfFileID, nil
	}

	req, err := http.NewRequestWithContext(ctx, http.MethodGet, url, nil)
	if err != nil {
		return "", err
	}

	resp, err := srv.HTTPClient.Do(req)
	if err != nil {
		return "", err
	}

	defer func() {
		_ = resp.Body.Close()
	}()

	if resp.StatusCode != http.StatusOK {
		return "", fmt.Errorf("%w: invalid response status %d: %s", grpcerr.ProviderError, resp.StatusCode, url)
	}

	fileName := path.Base(strings.TrimSuffix(url, "/"))

	mimeType, _, err := mime.ParseMediaType(resp.Header.Get("Content-Type"))
	if err != nil {
		switch path.Ext(fileName) {
		case ".pdf":
			mimeType = "application/pdf"
		case ".png":
			mimeType = "image/png"
		case ".jpg", ".jpeg":
			mimeType = "image/jpeg"
		case ".gif":
			mimeType = "image/gif"
		}
	}

	tmpdir := os.TempDir() + "/rentprofile"

	err = os.MkdirAll(tmpdir, 0755)
	if err != nil {
		return "", err
	}

	f, err := os.CreateTemp(tmpdir, "pdf-*")
	if err != nil {
		return "", err
	}

	defer func() {
		_ = f.Close()
		_ = os.Remove(f.Name())
	}()

	_, err = io.Copy(f, resp.Body)
	if err != nil {
		return "", err
	}

	_, err = f.Seek(0, io.SeekStart)
	if err != nil {
		return "", err
	}

	file, err := document.UploadFile(
		ctx,
		srv.Connections.Document,
		fileName,
		mimeType,
		&document.FileAttribute{
			Data: map[string]string{
				document.FileAttributeDocumentType.String(): document.TypeReferencingReport,
			},
		},
		f,
	)
	if err != nil {
		return "", err
	}

	return file.GetId(), nil
}

func (srv *serviceImpl) SkipChecks(ctx context.Context, typ model.PollingType, offerID string) error {
	return srv.Repository.PollingRemoveByTypeOffer(ctx, typ, offerID)
}

// getPDF tries to get rentprofile pdf URL, download report and then upload it into document service.
// Returns file ID on success and error otherwise.
func (srv *serviceImpl) getPDF(
	ctx context.Context,
	requestID string,
) (fileID string, err error) {
	var pdfURL string

	res, errr := srv.Client.Result(ctx, &rentprofile.ResultRequest{
		RequestID:   requestID,
		PDFRequired: rentprofile.BooleanTrue,
	})
	if errr != nil {
		return "", errr
	}

	pdfURL = res.PDFURL
	if pdfURL != "" {
		fileID, err = srv.downloadPDF(ctx, pdfURL)
	}

	return
}
