package service

import (
	"context"
	"net/url"
	"strings"
	"time"

	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	identity "gitlab.mashroom.com/mashroom/backend/identity/api"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/model"
	"gitlab.mashroom.com/mashroom/backend/identity/internal/rentprofile"
	offer "gitlab.mashroom.com/mashroom/backend/offer/api"
)

type serviceImpl struct {
	Config
}

func (srv *serviceImpl) emulatedTenantsInvite(
	req *rentprofile.InviteRequest,
) (resp *rentprofile.InviteResponse, data []*model.PollingEmulatedData) {
	var apps []*rentprofile.InviteResponseApplicant

	for _, a := range req.Applicants {
		apps = append(apps, &rentprofile.InviteResponseApplicant{
			Name:      a.Name,
			RequestID: a.Email,
		})

		var risk identity.TenantRisk

		var riskdoc string

		switch {
		case strings.Contains(a.Email, "-high@"):
			risk, riskdoc = identity.TenantRisk_HIGH_RISK, srv.Config.Emulated.HighRiskFileID
		case strings.Contains(a.Email, "-medium@"):
			risk, riskdoc = identity.TenantRisk_MEDIUM_RISK, srv.Config.Emulated.MediumRiskFileID
		case strings.Contains(a.Email, "@"):
			risk, riskdoc = identity.TenantRisk_LOW_RISK, srv.Config.Emulated.LowRiskFileID
		}

		data = append(data, &model.PollingEmulatedData{
			ProfileLink:  "dev.env.mashroom.com",
			RiskDocument: riskdoc,
			Risk:         risk,
		})
	}

	resp = &rentprofile.InviteResponse{
		Applicants: apps,
	}

	return
}

func riskToScore(r identity.TenantRisk) string {
	switch r {
	case identity.TenantRisk_HIGH_RISK:
		return "high"
	case identity.TenantRisk_MEDIUM_RISK:
		return "medium"
	case identity.TenantRisk_LOW_RISK:
		return "low"
	}

	return ""
}

func (srv *serviceImpl) TenantsInvite(
	ctx context.Context,
	req *identity.TenantsInviteRequest,
) (resp *identity.TenantsInviteResponse, err error) {
	var applicants []*rentprofile.InviteRequestApplicant

	for _, t := range req.GetTenants() {
		applicants = append(applicants, &rentprofile.InviteRequestApplicant{
			Name:   t.GetApplicantName(),
			Email:  t.GetApplicantEmail(),
			Mobile: t.GetApplicantMobile(),
		})
	}

	for _, t := range req.GetGuarantors() {
		applicants = append(applicants, &rentprofile.InviteRequestApplicant{
			Name:           t.GetApplicantName(),
			Email:          t.GetApplicantEmail(),
			Mobile:         t.GetApplicantMobile(),
			IsGuarantorFor: t.GetTenantName(),
		})
	}

	invreq := &rentprofile.InviteRequest{
		AddressLine1: req.GetAddress(),
		Postcode:     req.GetPostcode(),
		TotalRent:    int(req.GetTotalRent()),
		Applicants:   applicants,
	}

	var invresp *rentprofile.InviteResponse

	var data []*model.PollingEmulatedData

	if srv.Emulated.Enabled || grpcctx.ContextIsTestUser(ctx) {
		invresp, data = srv.emulatedTenantsInvite(invreq)
	} else {
		invresp, err = srv.Client.Invite(ctx, invreq)
	}

	if err != nil {
		return nil, err
	}

	var reqids []string

	for i, a := range invresp.Applicants {
		if a.RequestID == "" {
			// rentprofile behavior investigation: check logs for error below
			// we should continue iteration in this case, but cannot
			// due to response must contain items for each applicant
			log.WithFields(log.Fields{
				"offer": req.GetId(),
				"name":  a.Name,
			}).Errorf("got an empty renprofile request_id")
		}

		m := &model.Polling{
			Offer:        req.GetId(),
			Name:         a.Name,
			RequestID:    a.RequestID,
			CreatedAt:    time.Now(),
			LastPolledAt: time.Now(),
		}

		if data != nil {
			m.Emulated = data[i]
		}

		err = srv.Repository.PollingInsert(ctx, model.PollingTypeTenant, m)
		if err != nil {
			return nil, err
		}

		reqids = append(reqids, a.RequestID)
	}

	bgctx := grpcctx.ValuesBackgroundContext(ctx)

	if data != nil {
		go func() {
			time.Sleep(time.Second * 10)

			for i, a := range invresp.Applicants {
				_, err = srv.Connections.Offer.TenantCheckUpdate(bgctx, &offer.TenantCheckUpdateRequest{
					Offer:       req.GetId(),
					Check:       a.RequestID,
					PdfId:       data[i].RiskDocument,
					RiskScore:   riskToScore(data[i].Risk),
					Status:      rentprofile.TenantStatusComplete,
					ProfileLink: data[i].ProfileLink,
				})
				if err != nil {
					log.WithError(err).WithField("tenant", *a).Error("Updating emulated tenant")
				}
			}
		}()
	}

	return &identity.TenantsInviteResponse{
		RequestId: reqids,
	}, nil
}

func (srv *serviceImpl) TenantsStatus(
	ctx context.Context,
	req *identity.TenantsStatusRequest,
) (*identity.TenantsStatusResponse, error) {
	pol, _ := srv.Repository.PollingGetByTypeRequestID(ctx, model.PollingTypeTenant, req.GetRequestId())

	if pol != nil && pol.Emulated != nil {
		return &identity.TenantsStatusResponse{
			Status:       rentprofile.TenantStatusComplete,
			InviteDate:   pol.CreatedAt.Unix(),
			ProfileLink:  pol.Emulated.ProfileLink,
			Risk:         pol.Emulated.Risk,
			RiskDocument: pol.Emulated.RiskDocument,
		}, nil
	}

	if srv.Emulated.Enabled {
		return nil, errEmulateModeIsActive
	}

	logger := log.WithFields(log.Fields{
		"action": "TenantsStatus",
		"poll": func() string {
			if pol != nil {
				return pol.ID.Hex()
			}
			return "false"
		}(),
		"request_id": req.GetRequestId(),
	})

	err := srv.checkUpdate(ctx, req.RequestId, model.PollingTypeTenant, pol)

	switch {
	case err != nil:
		logger.WithError(err).Errorf("check status failed")
	case pol.PDFID == "":
		logger.Warn("pdf was not downloaded")
	}

	return &identity.TenantsStatusResponse{
		Status:       pol.Status,
		InviteDate:   pol.InviteDate.Unix(),
		ProfileLink:  pol.ProfileLink,
		Risk:         translateRiskScore(pol.RiskScore),
		RiskDocument: pol.PDFID,
	}, nil
}

func (srv *serviceImpl) SanctionsCheck(
	ctx context.Context,
	req *identity.SanctionsCheckRequest,
) (*identity.SanctionsCheckResponse, error) {
	count, err := srv.Repository.SanctionCountByName(ctx, letterGroups(req.GetName())...)
	if err != nil {
		return nil, err
	}

	return &identity.SanctionsCheckResponse{
		Found: count > 0,
	}, nil
}

func (srv *serviceImpl) DownloadPDF(
	ctx context.Context,
	req *identity.DownloadPDFRequest,
) (*identity.DownloadPDFResponse, error) {
	logger := log.WithFields(log.Fields{
		"action":     "DownloadPDF",
		"request_id": req.GetRequestId(),
	})

	fileID, err := srv.getPDF(ctx, req.GetRequestId())

	if err != nil {
		logger.WithError(err)
	} else {
		logger.Infof("fileID %s", fileID)
	}

	return &identity.DownloadPDFResponse{
		FileId: fileID,
	}, err
}

func (srv *serviceImpl) WebhookStatus(
	ctx context.Context,
	req *identity.WebhookStatusRequest,
) (*identity.WebhookStatusResponse, error) {
	var m rentprofile.StatusResponse

	values, err := url.ParseQuery(string(req.GetData()))
	if err != nil {
		return nil, err
	}

	m.RequestID = values.Get("request_id")
	m.Status = values.Get("status")
	m.InviteDate = values.Get("invite_date")
	m.ProfileLink = values.Get("profile_link")

	pol, err := srv.Repository.PollingGetByRequestID(ctx, m.RequestID)
	if err != nil {
		return nil, err
	}

	changed, err := srv.processStatus(ctx, pol, &m)
	if err != nil {
		return nil, err
	}

	err = srv.processChange(ctx, pol, changed, false)
	if err != nil {
		return nil, err
	}

	return &identity.WebhookStatusResponse{}, nil
}

func (srv *serviceImpl) WebhookResult(
	ctx context.Context,
	req *identity.WebhookResultRequest,
) (*identity.WebhookResultResponse, error) {
	var m rentprofile.ResultResponse

	values, err := url.ParseQuery(string(req.GetData()))
	if err != nil {
		return nil, err
	}

	m.RequestID = values.Get("request_id")
	m.PDFURL = values.Get("pdf_url")
	m.Name = values.Get("name")
	m.RiskScore = values.Get("risk_score")
	m.EmailAddress = values.Get("email_address")

	pol, err := srv.Repository.PollingGetByRequestID(ctx, m.RequestID)
	if err != nil {
		return nil, err
	}

	changed, err := srv.processResult(ctx, pol, &m)
	if err != nil {
		return nil, err
	}

	err = srv.processChange(ctx, pol, false, changed)
	if err != nil {
		return nil, err
	}

	return &identity.WebhookResultResponse{}, nil
}
